%===============================================================================
%NOME: condicoes.pl
%
%Projeto: Correo de Programas
%
%   Verso: V 1.00
%   Autora : Juliana Carpes Imperial - 0124811-7
%
%Descrio do mdulo
%   Objetivos:
%      - ler as pr e ps-condies, e os invariantes que sero utilizados
%        durante a correo do programa em ALP; 
%      - determinar se as pr e ps-condies, e os invariantes esto
%        sintaticamente corretos; 
%      - achar os loops existentes no programa escrito em ALP reconhecido, para
%        que seus invariantes possam ser dados pelo usurio;
%      - procurar o invariante de um determinado loop em uma lista de
%        < loop, invariante >;
%      - ver se uma pr-condio  equivalente a uma ps-condio aps um
%        comando de atribuio; e
%      - imprimir uma condio em formato HTML.
%===============================================================================


%===============================================================================
%Descrio da consulta: declara os operadores para as condies contidas nas pr
%                       e ps-condies, e nos invariantes que no so
%                       pr-definidos pelo SWI-Prolog.
%===============================================================================

:- op( 700, xfx, '<>' )  ,   % diferente
   op( 700, xfx, '<=' )  ,   % menor ou igual
   op( 699, xfy, '<=>' ) ,   % equivalncia
   op( 698, xfy, '=>' )  ,   % implicao
   op( 697, xfy, 'or' )  ,   % operador booleano or (\/)
   op( 696, xfy, 'and' ) ,   % operador booleando and (/\) 
   op( 695, fy, 'not' )  .   % operador booleando not (~)


%===============================================================================
%Nome: lerCondicoes
%Descrio do predicado: l as condies dadas pelo usurio para pr e ps-
%                        condies, e invariantes. 
%
%Varivel de entrada: 
%   - Tokens: a lista de tokens ALP lida.
%
%Variveis de sada:
%   - Invars:  a lista de invariantes para os loops do programa em ALP;
%   - PreCond: a pr-condio para o clculo de hoare; e
%   - PosCond: a ps-condio para o clculo de hoare.  
%===============================================================================

lerCondicoes( Tokens, Invars, PreCond, PosCond ) :-
   lerCondicao( PreCond, '\nPr-condio: ' )    ,  % l pr-condio
   lerCondicao( PosCond, '\nPs-condio: ' )    ,  % l ps-condio
   lerInvariantes( Tokens, Invars )              ,  % l invariantes
   nl                                            ,
   !                                             .
   

%===============================================================================
%Nome: lerCondicao
%Descrio do predicado: l uma condio dada pelo usurio para pr ou ps-
%                        condio, ou um invariante. 
%
%Varivel de entrada: 
%   - Pedido: instruo a ser impressa na tela pedindo um tipo de condio ao
%             usurio.
%
%Varivel de sada:
%   - Condicao: a condio dada pelo usurio. 
%===============================================================================

% pede um determinado tipo de condio ao usurio, a l e checa sua validade. Se
% for invlida, pede novamente
lerCondicao( Condicao, Pedido )                       :-
   write( Pedido )                                    ,
   ( ( read( Condicao )                               ,
       checarCondicao( Condicao ) )                   ;
     ( write( '\nCondicao invalida! Tente de novo!' ) ,
       lerCondicao( Condicao, Pedido ) ) )            .


%===============================================================================
%Nome: lerInvariantes
%Descrio do predicado: l os ivariantes para os loops encontrados no programa
%                        em ALP. 
%
%Varivel de entrada: 
%   - Tokens: a lista de tokens ALP lida.
%
%Varivel de sada:
%   - Invars: lista de invariantes, um para cada loop encontrado no programa. 
%===============================================================================

lerInvariantes( Tokens, Invars ) :-
   acharLoops( Tokens, Loops )   ,   % acha os loops do programa em ALP
   pedirInvars( Loops, Invars )  .   % l os invariantes para os loops


%===============================================================================
%Nome: checarCondicao
%Descrio do predicado: determina se uma pr ou ps-condio, ou um invariante
%                        est com a sintaxe correta (expresso booleana).
%
%Varivel de entrada: 
%   - Condicao: a pr ou ps-condio, ou o invariante que ter sua sintaxe
%               checada.
%===============================================================================

% se a condio representada dentro do Prolog for uma varivel livre (isso 
% ocorre se o usurio entra variveis comeadas com letras maisculas sem ' '),
% ela  considerada invlida
checarCondicao( Condicao ) :-
   var( Condicao )         ,
   !                       ,
   fail                    .

% testa se a condio  uma expresso booleana composta de conectivo binrio:
% se-somente-se (<=>), se-ento (=>), or ou and e checa recursivamente cada um
% dos operandos do conectivo recursivamente 
checarCondicao( Condicao )              :-
   ( Condicao = ( Cond_1 <=> Cond_2 )   ;
     Condicao = ( Cond_1 => Cond_2 )    ;
     Condicao = ( Cond_1 or Cond_2 )    ;
     Condicao = ( Cond_1 and Cond_2 ) ) ,
   checarCondicao( Cond_1 )             ,
   checarCondicao( Cond_2 )             .

% testa se a condio  uma condio negada (not Condicao) e checa 
% recursivamente a condio que vem depois do operador unrio not (Condicao)
checarCondicao( not Condicao ) :-
   checarCondicao( Condicao )  . 

% testa se a condio  um tomo
checarCondicao( Condicao ) :-
   atom( Condicao )        .

% testa se a condio  uma expresso booleana obtida atravs de uma comparao
% entre duas expresses aritmticas: =, <>, <, >, <= ou >= e checa 
% recursivamente cada um dos operandos do conectivo recursivamente 
checarCondicao( Condicao )           :-
   ( Condicao = ( Exp_1 = Exp_2 )    ;
     Condicao = ( Exp_1 <> Exp_2 )   ;
     Condicao = ( Exp_1 < Exp_2 )    ;
     Condicao = ( Exp_1 > Exp_2 )    ;
     Condicao = ( Exp_1 <= Exp_2 )   ;
     Condicao = ( Exp_1 >= Exp_2 ) ) ,
   checarExpressao( Exp_1 )          ,
   checarExpressao( Exp_2 )          .


%===============================================================================
%Nome: checarExpressao
%Descrio do predicado: determina se uma expresso aritmtica est com a 
%                        sintaxe correta (sintaxe igual a de um expressao 
%                        inteira em ALP).
%
%Varivel de entrada: 
%   - Expressao: a expresso aritmtica que ter sua sintaxe checada.
%===============================================================================

% se a expresso representada dentro do Prolog for uma varivel livre (isso
% ocorre se o usurio entra variveis comeadas com letras maisculas sem ' '),
% ela  considerada invlida
checarExpressao( Expressao ) :-
   var( Expressao )          ,
   !                         ,
   fail                      .

% testa se a expresso aritmtica  composta de operador binrio: +, -, * ou /
% e checa recursivamente cada um dos operandos da operao recursivamente 
checarExpressao( Expressao )         :-
   ( Expressao = ( Exp_1 + Exp_2 )   ;
     Expressao = ( Exp_1 - Exp_2 )   ;
     Expressao = ( Exp_1 * Exp_2 )   ;
     Expressao = ( Exp_1 / Exp_2 ) ) ,
   checarExpressao( Exp_1 )          ,
   checarExpressao( Exp_2 )          .       
             
% testa se a expresso aritmtica  uma expresso negada (- Expressao) e checa 
% recursivamente a expresso que vem depois do sinal de menos (Expressao)
checarExpressao( - Expressao )  :-  
   checarExpressao( Expressao ) .

% testa se a expresso  um tomo (varivel em ALP) ou um inteiro
checarExpressao( Expressao ) :-
   atom( Expressao )         ;
   integer( Expressao )      . 


%===============================================================================
%Nome: acharLoops
%Descrio do predicado: acha os loops existentes em uma lista de tokens lidos
%                        de um programa em ALP.
%
%Varivel de entrada: 
%   - Tokens: a lista de tokens ALP lida.
%
%Varivel de sada:
%   - Loops: os loops encontrados na lista de tokens ALP dada.
%
%Assertiva de entrada:
%   - A anlise sinttica do programa em ALP foi feita com sucesso.
%===============================================================================

% se todos os tokens j foram analisados, ento j foram encontados todos os 
% loops (possivelmente nenhum), no havendo mais nenhum a ser encontrado
acharLoops( [ ], [ ] ) .

% se um token while for encontado,  porque foi encontrado o incio de um loop,
% o qual ser lido at o final, e o resto dos tokens sero analisados para se
% saber se h algum outro loop
acharLoops( [ while | Tokens ], [ while( Teste, Corpo ) | Loops ] )           :-
   serComando( while( Teste, Corpo ), [ while | Tokens ], [ while | Tokens ],
                                                          _Resto )            ,
   acharLoops( Tokens, Loops )                                                .

% se um token qualquer for encontado, a busca por loops continua no prximo 
% token
acharLoops( [ _Token | Tokens ], Loops ) :-
   acharLoops( Tokens, Loops )           .


%===============================================================================
%Nome: pedirInvars
%Descrio do predicado: pede ao usurio os invariantes para os loops do
%                        programa em ALP.
%
%Varivel de entrada:
%   - Loops: a lista de loops do programa em ALP encontrados.
%
%Varivel de sada:
%   - Invars: a lista de invariantes para os loops.
%===============================================================================

% se todos os loops j foram percorridos, ento j foram lidos todos os 
% invariantes (possivelmente nenhum), no havendo mais nenhum invariante a ser
% lido
pedirInvars( [ ], [ ] ) .

% l o invariante do primeiro loop e, recursivamente, l os invariantes para os
% demais loops 
pedirInvars( [ Loop | Loops ], [ inv( Loop, Invar ) | Invars ] ) :-
   nl                                                            ,
   write( Loop )                                                 ,
   lerCondicao( Invar, '\nInvariante: ' )                        ,
   pedirInvars( Loops, Invars )                                  .


%===============================================================================
%Nome: procurarInvar
%Descrio do predicado: Procura o invariante para um dado loop.
%
%Variveis de entrada:
%   - Invars:                a lista de pares < loop, invariante >; e
%   - while( Teste, Corpo ): o loop a ter o invariante procurado.
%
%Varivel de sada:
%   - Invar: o invariante encontrado.
%===============================================================================

% se o par sendo analisado possui o loop cujo invariante  o que est sendo
% procurado, o invariante do par  a resposta
procurarInvar( [ inv( Loop, Invar ) | _Invars ], while( Teste, Corpo ),
                                                               Invar ) :-
   Loop == while( Teste, Corpo )                                       .

% caso contrrio, o invariante  procurado nos demais pares
procurarInvar( [ _LoopInv | Invars ], while( Teste, Corpo ), Invar ) :-
   procurarInvar( Invars, while( Teste, Corpo ), Invar )             .


%===============================================================================
%Nome: serIgual
%Descrio do predicado: analisa se uma pr-condio  equivalente a uma ps-
%                        condio aps um comando de atribuio.
%
%Variveis de entrada:
%   - PreCond: a pr-condio;
%   - Var:     a varivel que recebe um valor na atribuio Var := Exp;
%   - Exp:     a expresso que  atribuda na atribuio Var := Exp; e
%   - PosCond: a ps-condio.
%===============================================================================

% a pr-condio precisa ter a mesma forma de uma ps-condio para ser 
% equivalente a mesma aps um comando de atribuio, e este predicado verifica 
% se ambas so da forma A op B, onde A e B tambm devem ser checados 
% recursivamente para ver se possuem esta propriedade
serIgual( PreCond, Var, Exp, PosCond )  :-
   ( ( PosCond = ( Exp_1 <=> Exp_2 )    ,   % forma A <=> B
       PreCond = ( Exp_3 <=> Exp_4 ) )  ;
     ( PosCond = ( Exp_1 => Exp_2 )     ,   % forma A => B
       PreCond = ( Exp_3 => Exp_4 ) )   ;
     ( PosCond = ( Exp_1 or Exp_2 )     ,   % forma A or B
       PreCond = ( Exp_3 or Exp_4 ) )   ;
     ( PosCond = ( Exp_1 and Exp_2 )    ,   % forma A and B
       PreCond = ( Exp_3 and Exp_4 ) )  ;
     ( PosCond = ( Exp_1 = Exp_2 )      ,   % forma A = B
       PreCond = ( Exp_3 = Exp_4 ) )    ;
     ( PosCond = ( Exp_1 <> Exp_2 )     ,   % forma A <> B
       PreCond = ( Exp_3 <> Exp_4 ) )   ;
     ( PosCond = ( Exp_1 < Exp_2 )      ,   % forma A < B
       PreCond = ( Exp_3 < Exp_4 ) )    ;
     ( PosCond = ( Exp_1 > Exp_2 )      ,   % forma A > B
       PreCond = ( Exp_3 > Exp_4 ) )    ;
     ( PosCond = ( Exp_1 <= Exp_2 )     ,   % forma A <= B
       PreCond = ( Exp_3 <= Exp_4 ) )   ;
     ( PosCond = ( Exp_1 >= Exp_2 )     ,   % forma A >= B
       PreCond = ( Exp_3 >= Exp_4 ) )   ;
     ( PosCond = ( Exp_1 + Exp_2 )      ,   % forma A + B
       PreCond = ( Exp_3 + Exp_4 ) )    ;
     ( PosCond = ( Exp_1 - Exp_2 )      ,   % forma A - B
       PreCond = ( Exp_3 - Exp_4 ) )    ;
     ( PosCond = ( Exp_1 * Exp_2 )      ,   % forma A * B
       PreCond = ( Exp_3 * Exp_4 ) )    ;
     ( PosCond = ( Exp_1 / Exp_2 )      ,   % forma A / B
       PreCond = ( Exp_3 / Exp_4 ) ) )  ,
   serIgual( Exp_3, Var, Exp, Exp_1 )   ,   % anlise recursiva
   serIgual( Exp_4, Var, Exp, Exp_2 )   .   % anlise recursiva

% a pr-condio precisa ter a mesma forma de uma ps-condio para ser 
% equivalente a mesma aps um comando de atribuio, e este predicado verifica 
% se ambas so da forma not A, onde A tambm deve ser checado recursivamente 
% para ver se possui esta propriedade
serIgual( not PreCond, Var, Exp, not PosCond ) :-
   serIgual( PreCond, Var, Exp, PosCond )      . 

% a pr-condio precisa ter a mesma forma de uma ps-condio para ser
% equivalente a mesma aps um comando de atribuio, e este predicado verifica 
% se ambas so da forma - A, onde A tambm deve ser checado recursivamente para 
% ver se possui esta propriedade
serIgual( - PreCond, Var, Exp, - PosCond ) :-
   serIgual( PreCond, Var, Exp, PosCond )  .

% neste caso, a pr-condio  equivalente a ps-condio aps o comando de
% atribuio, com Exp sendo substitudo por var na pr-condio
serIgual( Exp, Var, Exp, Var ) .

% neste caso, a pr-condio  equivalente a ps-condio aps o comando de
% atribuio, porque so iguais e Var no aparece na ps-condio (se Var no
% aparece na ps-condio, nada tem que ser substitudo)
serIgual( Atomo, Var, _Exp, Atomo ) :-
   Atomo \== Var                    ,
   ( atom( Atomo )                  ;
     integer( Atomo ) )             .


%===============================================================================
%Nome: imprimirCond
%Descrio do predicado: imprime uma pr ou ps-condio, invariante ou sentena
%                        a ser demonstrada.
%
%Varivel de entrada:
%   Condio: a condio (pr ou ps-condio, invariante ou sentena) a ser
%             impressa.
%===============================================================================

% a condio a ser impressa  um tomo (varivel ou as constantes true ou false)
% ou um inteiro
imprimirCond( Condicao )   :-
   ( atom( Condicao )      ;
     integer( Condicao ) ) ,
   write( Condicao )       .

% a condio a ser impressa  da forma A op B, onde A e B so impressos
% recursivamente, sendo a expresso impressa com a devida parentizao
imprimirCond( Condicao )              :-
   ( ( Condicao = ( Exp_1 <=> Exp_2 ) ,   % forma A <=> B
       Op = ' <=> ' )                 ;
     ( Condicao = ( Exp_1 => Exp_2 )  ,   % forma A => B
       Op = ' => ' )                  ;
     ( Condicao = ( Exp_1 or Exp_2 )  ,   % forma A or B
       Op = ' or ' )                  ;
     ( Condicao = ( Exp_1 and Exp_2 ) ,   % forma A and B
       Op = ' and ' )                 ;
     ( Condicao = ( Exp_1 = Exp_2 )   ,   % forma A = B
       Op = ' = ' )                   ;
     ( Condicao = ( Exp_1 <> Exp_2 )  ,   % forma A <> B
       Op = ' <> ' )                  ;
     ( Condicao = ( Exp_1 < Exp_2 )   ,   % forma A < B
       Op = ' < ' )                   ;
     ( Condicao = ( Exp_1 > Exp_2 )   ,   % forma A > B
       Op = ' > ' )                   ;
     ( Condicao = ( Exp_1 <= Exp_2 )  ,   % forma A <= B
       Op = ' <= ' )                  ;
     ( Condicao = ( Exp_1 >= Exp_2 )  ,   % forma A >= B
       Op = ' >= ' )                  ;
     ( Condicao = ( Exp_1 + Exp_2 )   ,   % forma A + B
       Op = ' + ' )                   ;
     ( Condicao = ( Exp_1 - Exp_2 )   ,   % forma A - B
       Op = ' - ' )                   ;
     ( Condicao = ( Exp_1 * Exp_2 )   ,   % forma A * B
       Op = ' * ' )                   ;
     ( Condicao = ( Exp_1 / Exp_2 )   ,   % forma A / B
       Op = ' / ' ) )                 ,
   write( '( ' )                      ,
   imprimirCond( Exp_1 )              ,   % anlise recursiva
   write( Op )                        ,
   imprimirCond( Exp_2 )              ,   % anlise recursiva
   write( ' )' )                      .

% a condio a ser impressa  da forma - A, onde A  impresso recursivamente
imprimirCond( - Cond )  :-
   write( '- ' )        ,
   imprimirCond( Cond ) .

% a condio a ser impressa  da forma not A, onde A  impresso recursivamente
imprimirCond( not Cond ) :-
   write( 'not ' )       ,
   imprimirCond( Cond )  .
