%===============================================================================
%NOME: expressoes.pl
%
%Projeto: Correo de Programas
%
%   Verso: V 1.00
%   Autora : Juliana Carpes Imperial - 0124811-7
%
%Descrio do mdulo
%   Objetivo:
%      - converter uma expresso lida pelo analisador sinttico para o formato
%        das condies utilizado durante a correo de programas.
%===============================================================================


%===============================================================================
%Nome: converter
%Descrio do predicado: converte uma expresso lida pelo analisador sinttico
%                        para o formato das condies utilizado durante a
%                        correo de programas.
%
%Varivel de entrada:
%   - a expresso a ser convertida.
%
%Varivel de sada:
%   - a expresso convertida para o mesmo formato de uma condio.
%===============================================================================

converter( num( Num ), Num ) .  % converte um nmero
converter( ide( Var ), Var ) .  % converte uma varivel
converter( true, true )      .  % converte a constante booleana true
converter( false, false )    .  % converte a constante booleana false

% converte uma soma, convertendo recursivamente seus operandos
converter( expInt( soma, Exp_1, Exp_2 ), Exp_3 + Exp_4 ) :-
   converter( Exp_1, Exp_3 )                             ,
   converter( Exp_2, Exp_4 )                             .

% converte uma diferena, convertendo recursivamente seus operandos
converter( expInt( menos, Exp_1, Exp_2 ), Exp_3 - Exp_4 ) :-
   converter( Exp_1, Exp_3 )                              ,
   converter( Exp_2, Exp_4 )                              .

% converte uma multiplicao, convertendo recursivamente seus operandos
converter( expInt( mult, Exp_1, Exp_2 ), Exp_3 * Exp_4 ) :-
   converter( Exp_1, Exp_3 )                             ,
   converter( Exp_2, Exp_4 )                             .

% converte uma diviso, convertendo recursivamente seus operandos
converter( expInt( divide, Exp_1, Exp_2 ), Exp_3 / Exp_4 ) :-
   converter( Exp_1, Exp_3 )                               ,
   converter( Exp_2, Exp_4 )                               .

% converte uma inverso de sinal, convertendo recursivamente a expresso a ter o
% sinal invertido
converter( menos( Exp_1 ), - Exp_2 ) :-
   converter( Exp_1, Exp_2 )         .

% converte um and lgico, convertendo recursivamente seus operandos
converter( expBool( and, Exp_1, Exp_2 ), Exp_3 and Exp_4 ) :-
   converter( Exp_1, Exp_3 )                               ,
   converter( Exp_2, Exp_4 )                               .

% converte um or lgico, convertendo recursivamente seus operandos
converter( expBool( or, Exp_1, Exp_2 ), Exp_3 or Exp_4 ) :-
   converter( Exp_1, Exp_3 )                             ,
   converter( Exp_2, Exp_4 )                             .

% converte uma negao, convertendo recursivamente a expresso negada
converter( neg( Exp_1 ), not Exp_2 ) :-
   converter( Exp_1, Exp_2 )         .

% converte uma comparao de igualdade, convertendo recursivamente seus 
% operandos
converter( expBool( igual, Exp_1, Exp_2 ), Exp_3 = Exp_4 ) :-
   converter( Exp_1, Exp_3 )                               ,
   converter( Exp_2, Exp_4 )                               .

% converte uma comparao de desigualdade, convertendo recursivamente seus 
% operandos
converter( expBool( difer, Exp_1, Exp_2 ), Exp_3 <> Exp_4 ) :-
   converter( Exp_1, Exp_3 )                                ,
   converter( Exp_2, Exp_4 )                                .

% converte uma comparao de maior, convertendo recursivamente seus operandos
converter( expBool( maior, Exp_1, Exp_2 ), Exp_3 > Exp_4 ) :-
   converter( Exp_1, Exp_3 )                               ,
   converter( Exp_2, Exp_4 )                               .

% converte uma comparao de menor, convertendo recursivamente seus operandos
converter( expBool( menor, Exp_1, Exp_2 ), Exp_3 < Exp_4 ) :-
   converter( Exp_1, Exp_3 )                                ,
   converter( Exp_2, Exp_4 )                                .

% converte uma comparao de maior ou igual, convertendo recursivamente seus 
% operandos
converter( expBool( maiorIg, Exp_1, Exp_2 ), Exp_3 >= Exp_4 ) :-
   converter( Exp_1, Exp_3 )                                  ,
   converter( Exp_2, Exp_4 )                                  .

% converte uma comparao de menor ou igual, convertendo recursivamente seus 
% operandos
converter( expBool( menorIg, Exp_1, Exp_2 ), Exp_3 <= Exp_4 ) :-
   converter( Exp_1, Exp_3 )                                  ,
   converter( Exp_2, Exp_4 )                                  .