%===============================================================================
%NOME: hoare.pl
%
%Projeto: Correo de Programas
%
%   Verso: V 1.00
%   Autora : Juliana Carpes Imperial - 0124811-7
%
%Descrio do mdulo
%   Objetivos:
%      - implementao das regras de inferncia do clculo de Hoare para fazer
%        a correo de programas escritos em ALP;
%      - implementao de um filtro que retira provas de programas que possuem
%        o conjunto de sentenas a ser provado igual a de outras; e
%      - implementao de predicados para criar uma pgina HTML contendo vrias
%        provas de programa utilizando o clculo de Hoare juntamente com as
%        sentenas que precisam ser demonstradas para que as provas sejam
%        vlidas.
%===============================================================================


%===============================================================================
%Nome: fazerHoare
%Descrio do predicado: aplica as regras de inferncia do clculo de hoare para
%                        fazer a correo do programa em ALP.
%
%Variveis de entrada:
%   - Invars:   a lista de pares < loop, invariante >;
%   - PreCond:  a pr-condio para o trecho de programa sendo corrigido;
%   - Cmds:     o trecho de programa sendo corrigido;
%   - PosCond:  a ps-condio para o trecho de programa sendo corrigido; e
%   - SentsAnt: sentenas a serem demonstradas antes de aplicar a regra.
%
%Variveis de sada:
%   - Prova:    a rvore de prova resultante da aplicao da regra; e
%   - SentsDep: sentenas a serem demonstradas depois da aplicao da regra.
%===============================================================================

% comando skip; regra  axioma: ________________
%                               { P } skip { P }
fazerHoare( _Invars, Cond, [ skip ], Cond, Prova, Sents, Sents ) :-
   Prova = [ [ Cond, [ skip ], Cond ] ]                          .

% comando de atribuio; regra  axioma: _________________________
%                                        { P[ x/E ] } x := E { P }
% pelo menos uma das condies deve ser totalmente ligada para que a regra
% funcione corretamente. A expresso no programa deve ser convertida para o
% formato de condio antes de tentar se chegar no axioma
fazerHoare( _Invars, PreCond, [ atribui( Var, Exp ) ], PosCond, Prova, Sents,
                                                                Sents ) :-
   ( checarCondicao( PreCond )                                          ;
     checarCondicao( PosCond ) )                                        ,
   converter( Exp, EquivExp )                                           ,
   serIgual( PreCond, Var, EquivExp, PosCond )                          ,
   Prova = [ [ PreCond, [ atribui( Var, Exp ) ], PosCond ] ]            .

% regra para mais de um comando: { P } CMD1 { R }  { R } CDM2 { Q }
%                                __________________________________
%                                     { P } CMD1 ; CMD2 { Q }
% deve haver mais de um comando para que a regra funcione corretamente. So
% tentadas duas alternativas, uma comeando pelo primeiro comando e outra
% pelos demais comandos
fazerHoare( Invars, PreCond, [ Cmd | Cmds ], PosCond, Prova, SentsAnt,
                                                             SentsDep )       :-
   length( Cmds, TamCmds )                                                    ,
   TamCmds > 0                                                                ,
   ( ( fazerHoare( Invars, PreCond, [ Cmd ], CondTemp, ProvaFilha_1, SentsAnt ,
                                                       SentsTemp )            ,
       fazerHoare( Invars, CondTemp, Cmds, PosCond, ProvaFilha_2, SentsTemp,
                                                    SentsDep ) )              ;
     ( fazerHoare( Invars, CondTemp, Cmds, PosCond, ProvaFilha_2, SentsAnt,
                                                    SentsTemp )               ,
       fazerHoare( Invars, PreCond, [ Cmd ], CondTemp, ProvaFilha_1, SentsTemp,
                                                       SentsDep ) ) )         ,
   Prova = [ [ PreCond, [ Cmd | Cmds ], PosCond ], ProvaFilha_1,
                                                   ProvaFilha_2 ]             .

% regra do if sem else: { P and B } CMD { Q }  P and not B => Q
%                       _______________________________________
%                            { P } if B then CMD fi { Q }
% o teste do if deve ser convertido para o formato de condio antes da
% tentativa de aplicar a regra. A ps-condio no deve ser varivel livre
% depois da chamada recursiva para que a sentena a ser demonstrada seja bem
% formada. Se a pr-condio for varivel livre, tenta-se dois chutes com os
% dois possveis casos do if. A sentena a ser demonstrada no if  includa no
% conjunto de sentenas.
fazerHoare( Invars, PreCond, [ if( Teste, Entao ) ], PosCond, Prova, SentsAnt,
                                                              SentsDep )    :-
   converter( Teste, EquivTeste )                                           ,
   fazerHoare( Invars, PreCond and EquivTeste, Entao, PosCond, ProvaFilha,
                                                      SentsAnt, SentsTemp ) ,
   nonvar( PosCond )                                                        ,
   SentsDep = [ ( PreCond and not EquivTeste ) => PosCond | SentsTemp ]     ,
   Prova = [ [ PreCond, [ if( Teste, Entao ) ], PosCond ], ProvaFilha,
             [ ( PreCond and not EquivTeste ) => PosCond ] ]                ,
   ( var( PreCond )                                                         ->
     ( PreCond = EquivTeste                                                 ;
       PreCond = not EquivTeste                                             ;
       true )                                                               ;
     true )                                                                 .

% regra do if com else: { P and B } CMD1 { Q }  { P and not B } CMD2 { Q }
%                       __________________________________________________
%                         { P } if B then CMD1 else CMD2 esle_fi { Q }
% o teste do if deve ser convertido para o formato de condio antes da
% tentativa de aplicar a regra. Se a pr-condio for varivel livre, tenta-se
% dois chutes com os dois possveis casos do if.
fazerHoare( Invars, PreCond, [ if( Teste, Entao, Senao ) ], PosCond, Prova,
                                                 SentsAnt, SentsDep )       :-
   converter( Teste, EquivTeste )                                           ,
   fazerHoare( Invars, PreCond and EquivTeste, Entao, PosCond, ProvaFilha_1,
                                                      SentsAnt, SentsTemp ) ,
   fazerHoare( Invars, PreCond and not EquivTeste, Senao, PosCond,
                       ProvaFilha_2, SentsTemp, SentsDep )                  ,
   Prova = [ [ PreCond, [ if( Teste, Entao, Senao ) ], PosCond ],
             ProvaFilha_1, ProvaFilha_2 ]                                   ,
   ( var( PreCond )                                                         ->
     ( PreCond = EquivTeste                                                 ;
       PreCond = not EquivTeste                                             ;
       true )                                                               ;
     true )                                                                 .

% regra do while:          { P and B } CMD { P }
%                 _______________________________________
%                 { P } while B do CMD od { P and not B }
% o teste do while deve ser convertido para o formato de condio antes da
% tentativa de aplicar a regra.
fazerHoare( Invars, Invar, [ while( Teste, Corpo ) ], Invar and not EquivTeste,
                                           Prova, SentsAnt, SentsDep )     :-
   converter( Teste, EquivTeste )                                          ,
   fazerHoare( Invars, Invar and EquivTeste, Corpo, Invar, ProvaFilha,
                                                    SentsAnt, SentsDep )   ,
   Prova = [ [ Invar, [ while( Teste, Corpo ) ], Invar and not EquivTeste ],
             ProvaFilha ]                                                  .

% neste caso, a pr-condio  o invariante (P), mas a ps precisa ser
% fortalecida para ser igual a P and not B, para que a regra do while escrita
% acima possa ser aplicada. O teste do while deve ser convertido para o formato
% de condio antes da recurso e a ps-condio no pode ser varivel livre
% para que a sentena a ser demonstrada seja bem formada. Esta  includa no
% conjunto de sentenas.
fazerHoare( Invars, Invar, [ while( Teste, Corpo ) ], PosCond, Prova, SentsAnt,
                                                               SentsDep )  :-
   converter( Teste, EquivTeste )                                          ,
   PosCond \== ( Invar and not EquivTeste )                                ,
   nonvar( PosCond )                                                       ,
   fazerHoare( Invars, Invar, [ while( Teste, Corpo ) ],
               Invar and not EquivTeste, ProvaFilha, SentsAnt, SentsTemp ) ,
   SentsDep = [ ( Invar and not EquivTeste ) => PosCond | SentsTemp ]      ,
   Prova = [ [ Invar, [ while( Teste, Corpo ) ], PosCond ], ProvaFilha,
             [ ( Invar and not EquivTeste ) => PosCond ] ]                 .

% neste caso, a pr-condio no  o invariante (P), que antes precisa ser
% encontrado na lista de < loop, invariante > para que possa ser enfraquecido
% para ser igual a P, para que a regra do while escrita acima possa ser aplicada
% caso a ps tambm j esteja no formato correto. A pr-condio no pode ser
% varivel livre para que a sentena a ser demonstrada seja bem formada. Esta 
% includa no conjunto de sentenas.
fazerHoare( Invars, PreCond, [ while( Teste, Corpo ) ], PosCond, Prova,
                                                        SentsAnt, SentsDep ) :-
   procurarInvar( Invars, while( Teste, Corpo ), Invar )                     ,
   !                                                                         ,
   PreCond \== Invar                                                         ,
   ( var( PreCond )                                                          ->
     PreCond = Invar                                                         ;
     true )                                                                  ,
   fazerHoare( Invars, Invar, [ while( Teste, Corpo ) ], PosCond, ProvaFilha,
                                              SentsAnt, SentsTemp )          ,
   ( PreCond \== Invar                                                       ->
     ( SentsDep = [ PreCond => Invar | SentsTemp ]                           ,
       Prova = [ [ PreCond, [ while( Teste, Corpo ) ], PosCond ],
                 [ PreCond => Invar ], ProvaFilha ] )                        ;
     ( SentsDep = SentsTemp                                                  ,
       Prova = [ [ PreCond, [ while( Teste, Corpo ) ], PosCond ],
                 ProvaFilha ] ) )                                            . 

% fortalecimento da ps-condio: { P } CMD { R }  R => Q
%                                 _______________________
%                                     { P } CMD { Q }
% as pr e ps-condies no podem ser variveis livres, assim como a nova ps-
% condio aps a chamada recursiva, que tambm deve ser diferente da antiga
% ps-condio. A sentena a ser demonstrada no fortalecimento  includa no
% conjunto de sentenas.
fazerHoare( Invars, PreCond, [ Comando ], PosCond, Prova, SentsAnt,
                                                          SentsDep ) :-
   nonvar( PreCond )                                                 ,
   nonvar( PosCond )                                                 ,
   fazerHoare( Invars, PreCond, [ Comando ], NovaPosCond, ProvaFilha,
                                             SentsAnt, SentsTemp )   ,
   nonvar( NovaPosCond )                                             ,
   NovaPosCond \== PosCond                                           ,
   SentsDep = [ NovaPosCond => PosCond | SentsTemp ]                 ,
   Prova = [ [ PreCond, [ Comando ], PosCond ], ProvaFilha,
             [ NovaPosCond => PosCond ] ]                            .

% enfraquecimento da pr-condio: P => R  { R } CMD { Q }
%                                  _______________________
%                                      { P } CMD { Q }
% as pr e ps-condies no podem ser variveis livres, assim como a nova pr-
% condio aps a chamada recursiva, que tambm deve ser diferente da antiga
% pr-condio. A sentena a ser demonstrada no enfraquecimento  includa no
% conjunto de sentenas.
fazerHoare( Invars, PreCond, [ Comando ], PosCond, Prova, SentsAnt,
                                                          SentsDep )       :-
   nonvar( PreCond )                                                       ,
   nonvar( PosCond )                                                       ,
   fazerHoare( Invars, NovaPreCond, [ Comando ], PosCond, ProvaFilha,
                                                 SentsAnt, SentsTemp )     ,
   nonvar( NovaPreCond )                                                   ,
   NovaPreCond \== PreCond                                                 ,
   SentsDep = [ PreCond => NovaPreCond | SentsTemp ]                       ,
   Prova = [ [ PreCond, [ Comando ], PosCond ], [ PreCond => NovaPreCond ] ,
             ProvaFilha ]                                                  .


%===============================================================================
%Nome: retirarProvas
%Descrio do predicado: retira provas da lista de respostas (uma lista de
%                        < provas, sentenas a serem demonstradas > de forma
%                        que todas possuam um conjunto de sentenas a serem
%                        demonstradas diferentes entre si.
%
%Variveis de entrada:
%   - Respostas:  uma lista de < provas, sentenas a serem demonstradas >; e
%   - ListaSents: a lista conjunto de sentenas encontradas at ento.
%
%Varivel de sada:
%   - NovasResp: uma lista de < provas, sentenas a serem demonstradas > tal que
%                todas as provas possuem um conjunto de sentenas a serem
%                demonstradas diferentes entre si
%===============================================================================

% caso base, todas as provas j foram analisadas com relao s suas sentenas
retirarProvas( [ ], [ ], _ListaSents ) .

% o conjunto de sentenas da prova no  igual ao de nenhuma das provas j
% analisadas. A chamada recursiva coloca o conjunto desta para as prximas
% anlises.
retirarProvas( [ [ Prova, Sents ] | Respostas ],
               [ [ Prova, Sents ] | NovasResp ], ListaSents )    :-
   \+ member( Sents, ListaSents )                                ,
   retirarProvas( Respostas, NovasResp, [ Sents | ListaSents ] ) .

% o conjunto de sentenas da prova  igual ao de uma das provas j analisadas.
% Esta prova no far parte do conjunto de provas resultante.
retirarProvas( [ _Resposta | Respostas ], NovasResp, ListaSents ) :-
   retirarProvas( Respostas, NovasResp, ListaSents )              .

   
%===============================================================================
%Nome: imprimirSaida
%Descrio do predicado: cria uma pgina HTML contendo vrias provas de programa
%                        utilizando o clculo de Hoare juntamente com as
%                        sentenas que precisam ser demonstradas para que as
%                        provas sejam vlidas.
%
%Varivel de entrada:
%   - Respostas: a lista de < provas, sentenas a serem demonstradas > com todas
%                as provas possuindo um conjunto de sentenas a serem
%                demonstradas diferentes entre si.
%
%Assertiva de sada:
%   - a pgina HTML foi criada com sucesso e seu arquivo foi fechado, ou foi
%     detectado um erro de abertura, escrita ou fechamento de arquivo o que leva
%     o programa a ser abortado.
%===============================================================================

% tenta abrir o arquivo da pgina HTML para escrita. Se der erro, este 
% tratado. Se o arquivo for aberto com sucesso, seu cabealho  escrito,
% juntamente com as provas e seus conjuntos de sentenas devidamente formatados,
% e, por fim, so escritas as tags do fim da pgina. Se todo este processo de
% escrita no arquivo for feito com sucesso, ele  fechado. Se der algum erro de
% escrita ou fechamento, este  tratado
imprimirSaida( Respostas )                                                    :-
   catch( tell( 'saida.html' ), _Excecao, tratarErro( abertura ) )            ,
   catch( write( '<HTML>\n<HEAD> ' ), _Excecao, tratarErro( escrita ) )       ,
   catch( write( '<TITLE> Prova de ' ), _Excecao, tratarErro( escrita ) )     ,
   catch( write( 'Correo de Programas' ), _Excecao, tratarErro( escrita ) ) ,
   catch( write( ' </TITLE> </HEAD>\n' ), _Excecao, tratarErro( escrita ) )   ,
   catch( write( '<BODY> <B> <PRE>\n' ), _Excecao, tratarErro( escrita ) )    ,
   catch( imprimirProvas( Respostas ), _Excecao, tratarErro( escrita ) )      ,
   catch( write( '\n</B> </PRE> </BODY>' ), _Excecao, tratarErro( escrita ) ) ,
   catch( write( '\n<HTML>' ), _Excecao, tratarErro( escrita ) )              ,
   catch( told, _Excecao, tratarErro( fechamento ) )                          .


%===============================================================================
%Nome: imprimirProvas
%Descrio do predicado: imprime as provas e suas sentenas a serem demonstradas
%                        com formatao HTML.
%
%Varivel de entrada:
%   - Respostas: a lista de < provas, sentenas a serem demonstradas > com todas
%                as provas possuindo um conjunto de sentenas a serem
%                demonstradas diferentes entre si.
%
%Assertiva de entrada:
%   - O arquivo da pgina HTML est aberto.
%===============================================================================

% todas as provas e seus conjuntos de sentenas j foram impressos
imprimirProvas( [ ] ) .

% imprime uma prova em clculo de hoare e o conjunto de sentenas a ser
% demonstrado dela caso no seja vazio
imprimirProvas( [ [ Prova, Sents ] | Respostas ] )        :-
   write( '<H4>Prova: </H4>\n' )                          ,
   imprimirHoare( Prova, 1 )                              ,
   ( Sents \== [ ]                                        ->
     ( write( '\n<H4>Sentenas a serem provadas: </H4>' ) ,
       imprimirSents( Sents ) )                           ;
     true )                                               ,
   ( Respostas \== [ ]                                    ->
     write( '\n\n<HR>' )                                  ;
     true )                                               ,
   imprimirProvas( Respostas )                            .


%===============================================================================
%Nome: imprimirHoare
%Descrio do predicado: imprime as provas de correo de programa com
%                        formatao HTML.
%
%Variveis de entrada:
%   - Provas: a rvore com a prova de correo de programa; e
%   - Nivel:  o nvel que est sendo impresso a partir da raiz da rvore.
%
%Assertiva de entrada:
%   - O arquivo da pgina HTML est aberto.
%===============================================================================

% est se imprimindo um axioma, que  um trecho de prova sem filhos (uma folha
% da rvore), o que inclui as regras do skip e de atribuio
imprimirHoare( [ ProvaPai ], Nivel ) :-
   imprimirProva( ProvaPai, Nivel )  .

% est se imprimindo um ramo de prova que s tem um filho (while). O nvel de um
% filho  sempre maior do que o do pai em um
imprimirHoare( [ ProvaPai, ProvaFilha ], Nivel ) :-
   imprimirProva( ProvaPai, Nivel )              ,
   imprimirHoare( ProvaFilha, Nivel + 1 )        .

% est se imprimindo um ramo de prova que s tem dois filhos (regras do if,
% enfraquecimento de pr-condio, fortalecimento de ps-condio e para mais de
% um comando). O nvel de um filho  sempre maior do que o do pai em um
imprimirHoare( [ ProvaPai, ProvaFilha_1, ProvaFilha_2 ], Nivel ) :-
   imprimirProva( ProvaPai, Nivel )                              ,
   imprimirHoare( ProvaFilha_1, Nivel + 1 )                      ,
   imprimirHoare( ProvaFilha_2, Nivel + 1 )                      .


%===============================================================================
%Nome: imprimirProva
%Descrio do predicado: imprime um ramo de prova de correo de programa com
%                        formatao HTML.
%
%Variveis de entrada:
%   - Ramo:  um ramo da rvore de uma prova de correo de programa; e
%   - Nivel: o nvel que est sendo impresso a partir da raiz da rvore.
%
%Assertiva de entrada:
%   - O arquivo da pgina HTML est aberto.
%===============================================================================

% o ramo  um trecho se programa sendo corrigido da forma { P } CMD { Q }
imprimirProva( [ PreCond, Prog, PosCond ], Nivel ) :-
   tab( ( Nivel - 1 ) * 3 )                        ,
   write( '<FONT COLOR = "red">{ ' )               ,
   imprimirCond( PreCond )                         ,
   write( ' } </FONT>' )                           ,
   write( Prog )                                   ,
   write( '<FONT COLOR = "red"> { ' )              ,
   imprimirCond( PosCond )                         ,
   write( ' }</FONT>\n' )                          .

% o ramo  uma sentena a ser demonstrada da forma P => Q
imprimirProva( Condicao, Nivel )     :-
   Condicao = ( _Cond_1 => _Cond_2 ) ,
   tab( ( Nivel - 1 ) * 3 )          ,
   write( '<FONT COLOR = "blue">' )  ,
   imprimirCond( Condicao )          ,
   write( ' </FONT>' )               ,
   nl                                .


%===============================================================================
%Nome: imprimirSents
%Descrio do predicado: imprime uma lista de sentenas, uma em cada linha.
%
%Varivel de entrada:
%   - Sents: a lista de sentenas a ter seus elementos impressos.
%===============================================================================

imprimirSents( [ ] ) .                 % a lista j foi totalmente impressa

imprimirSents( [ Sent | Sents ] ) :-
   nl                             ,    % pula linha
   imprimirCond( Sent )           ,    % imprime uma sentena
   imprimirSents( Sents )         .    % imprime o resto da lista recursivamente
