%===============================================================================
%NOME: lexica.pl
%
%Projeto: Correo de Programas
%
%   Verso: V 1.00
%   Autora : Juliana Carpes Imperial - 0124811-7
%
%Descrio do mdulo
%   Objetivo:
%      - implementao dos predicados que fazem a anlise lxica da linguagem 
%        ALP.
%===============================================================================


%===============================================================================
%Nome: fazerAnaLexica
%Descrio do predicado: predicado principal do mdulo, o qual realiza todas as
%                        chamadas necessrias para realizar a anlise lxica.
%
%Varivel de sada: 
%   - Tokens: a lista de tokens do programa em ALP.
%
%Assertiva de sada:
%   - o arquivo do programa em ALP lido est fechado.
%===============================================================================

% l o nome de um arquivo e tenta abri-lo. Se der erro, este  tratado; se o 
% arquivo for aberto com sucesso, os tokens do arquivo ALP so lidos e se a 
% leitura dos tokens foi feita com sucesso, tenta fechar o arquivo. Se der erro,
% este  tratado, se no, exibe os tokens lidos ao usurio.
fazerAnaLexica( Tokens )                                                   :- 
   write( '\nEscaneando: anlise lxica...\n' )                            , 
   write( 'Entre nome do arquivo fonte entre \' \' e seguido de ponto: ' ) ,  
   read( Arquivo )                                                         , 
   catch( see( Arquivo ), _Excecao, tratarErro( abertura ) )               ,
   nl                                                                      , 
   escanear( Tokens )                                                      ,  
   catch( seen, _Excecao, tratarErro( fechamento ) )                       ,
   write( '\nAnlise lxica feita com sucesso.\n' )                        , 
   write( '\ntokens = ' )                                                  ,
   write( Tokens )                                                         , 
   nl                                                                      ,
   !                                                                       .


%===============================================================================
%Nome: serEspaco
%Descrio do predicado: determina se um caractere lido  o espao, ou seja, se 
%                        seu cdigo ASCII  32.
%
%Varivel de entrada: 
%   - o cdigo ASCII do caractere lido.
%===============================================================================

serEspaco( 32 ) .     


%===============================================================================
%Nome: serTab
%Descrio do predicado: determina se um caractere lido  o tab, ou seja, se seu 
%                        cdigo ASCII  9.
%
%Varivel de entrada: 
%   - o cdigo ASCII do caractere lido.
%===============================================================================

serTab( 9 ) .          

     
%===============================================================================
%Nome: serFimLinha
%Descrio do predicado: determina se um caractere lido  o que determina fim de 
%                        linha, ou seja, se seu cdigo ASCII  10. 
%
%Varivel de entrada: 
%   - o cdigo ASCII do caractere lido.
%===============================================================================

serFimLinha( 10 ) .   


%===============================================================================
%Nome: serFimArq
%Descrio do predicado: determina se um caractere lido  o que determina fim de 
%                        arquivo, ou seja, se seu cdigo ASCII  26 ou -1.
%
%Varivel de entrada: 
%   - o cdigo ASCII do caractere lido. 
%===============================================================================

serFimArq( 26 ) .   
serFimArq( -1 ) .


%===============================================================================
%Nome: serPonto
%Descrio do predicado: determina se um caractere lido  o ponto ("."), ou 
%                        seja, se seu cdigo ASCII  46.
%
%Varivel de entrada: 
%   - o cdigo ASCII do caractere lido.
%===============================================================================

serPonto( 46 ) .


%===============================================================================
%Nome: serSimples
%Descrio do predicado: determina se um caractere lido  tal que, se for o 
%                        primeiro de um token,  o prprio token.
%
%Varivel de entrada: 
%   - o cdigo ASCII do caractere lido.
%
%Varivel de sada:
%   - o nome do token correspondente ao caractere lido.     
%===============================================================================

serSimples( 40, parenEsq ) .   % caractere: ( -- token: parenEsq
serSimples( 41, parenDir ) .   % caractere: ) -- token: parenDir
serSimples( 42, mult )     .   % caractere: * -- token: mult
serSimples( 43, soma )     .   % caractere: + -- token: soma
serSimples( 45, menos )    .   % caractere: - -- token: menos
serSimples( 47, divide )   .   % caractere: / -- token: divide
serSimples( 59, ptEVirg )  .   % caractere: ; -- token: ptEVirg
serSimples( 61, igual )    .   % caractere: = -- token: igual


%===============================================================================
%Nome: serDuplo
%Descrio do predicado: determina se um caractere lido  tal que, se for o 
%                        primeiro de um token, ou  o prprio token ou  o 
%                        primeiro de um token com dois caracteres.
%
%Varivel de entrada: 
%   - o cdigo ASCII do caractere lido.
%
%Varivel de sada:
%   - o nome do token correspondente ao caractere lido se o primeiro s contiver
%     um caractere.     
%===============================================================================

serDuplo( 58, doisPts ) .   % caractere: : -- token: doisPts
serDuplo( 60, menor )   .   % caractere: < -- token: menor
serDuplo( 62, maior )   .   % caractere: > -- token: maior


%===============================================================================
%Nome: serPar
%Descrio do predicado: determina se dois caracteres lidos so tais que, se 
%                        forem lidos em seqncia, formam um token de dois
%                        caracteres.
%
%Variveis de entrada: 
%   - o cdigo ASCII do primeiro caractere lido; e
%   - o cdigo ASCII do segundo caractere lido.
%
%Varivel de sada:
%   - o nome do token correspondente ao par de caracteres lidos.
%===============================================================================

serPar( 58, 61, atribui ) .   % par de caracteres: := -- token: atribui
serPar( 60, 61, menorIg ) .   % par de caracteres: <= -- token: menorIg
serPar( 62, 61, maiorIg ) .   % par de caracteres: >= -- token: maiorIg
serPar( 60, 62, difer )   .   % par de caracteres: <> -- token: difer


%===============================================================================
%Nome: serReserv
%Descrio do predicado: determina se um identificador corresponde a uma palavra 
%                        reservada da linguagem ALP.                         
%
%Varivel de entrada: 
%   - o nome do identificador. 
%===============================================================================

serReserv( skip )    .   % palavra reservada skip (comando vazio)
serReserv( if )      .   % palavra reservada if (teste)  
serReserv( then )    .   % palavra reservada then (caso positivo do teste) 
serReserv( fi )      .   % palavra reservada fi (fim de if-then)
serReserv( else )    .   % palavra reservada else (caso negativo do teste)
serReserv( esle_fi ) .   % palavra reservada esle_fi (fim de if-then-else)
serReserv( while )   .   % palavra reservada while (repetio)
serReserv( do )      .   % palavra reservada do (o que faz a repetio)
serReserv( od )      .   % palavra reservada od (fim de while)
serReserv( and )     .   % palavra reservada and (conectivo booleano)
serReserv( or )      .   % palavra reservada or (conectivo booleano)
serReserv( not )     .   % palavra reservada not (conectivo booleano) 
serReserv( true )    .   % palavra reservada true (booleano primitivo)
serReserv( false )   .   % palavra reservada false (booleano primitivo) 


%===============================================================================
%Nome: serBranco
%Descrio do predicado: determina se um caractere lido  um branco.
%
%Varivel de entrada: 
%   - Car: o cdigo ASCII do caractere lido.
%===============================================================================

serBranco( Car )      :-  % um caractere  um branco se:
   serEspaco( Car )   ;   %  um espao, ou
   serTab( Car )      ;   %  um tab, ou
   serFimLinha( Car ) .   %  um fim de linha


%===============================================================================
%Nome: serMinuscula
%Descrio do predicado: determina se um caractere lido corresponde a uma letra
%                        minscula.
%
%Varivel de entrada: 
%   - Car: o cdigo ASCII do caractere lido.
%===============================================================================

serMinuscula( Car ) :-  % um caractere  uma letra minscula se:
   Car >= 97        ,   %  'a' (ASCII = 97) ou vem depois dele e
   Car =< 122       .   %  'z' (ASCII = 122) ou vem antes dele    
 

%===============================================================================
%Nome: serMaiuscula
%Descrio do predicado: determina se um caractere lido corresponde a uma letra
%                        maiscula.
%
%Varivel de entrada: 
%   - Car: o cdigo ASCII do caractere lido.
%===============================================================================

serMaiuscula( Car ) :-  % um caractere  uma letra maiscula se:
   Car >= 65        ,   %  'A' (ASCII = 65) ou vem depois dele e
   Car =< 90        .   %  'Z' (ASCII = 90) ou vem antes dele        


%===============================================================================
%Nome: serDigito 
%Descrio do predicado: determina se um caractere lido corresponde a um 
%                        dgito decimal.
%
%Varivel de entrada: 
%   - Car: o cdigo ASCII do caractere lido.
%===============================================================================

serDigito( Car ) :-  % um caractere  um dgito decimal se:        
   Car >= 48     ,   %  '0' (ASCII = 48) ou vem depois dele e
   Car =< 57     .   %  '9' (ASCII = 57) ou vem antes dele    


%===============================================================================
%Nome: serCarVal
%Descrio do predicado: determina se um caractere lido  vlido dentro de um 
%                        identificador.
%
%Varivel de entrada: 
%   - Car: o cdigo ASCII do caractere lido.
%===============================================================================

serCarVal( Car )       :-  % um caractere  vlido dentro de um identificador se                    
   serMinuscula( Car ) ;   %  uma letra minscula, ou
   serMaiuscula( Car ) ;   %  uma letra maiscula, ou
   serDigito( Car )    ;   %  um dgito decimal, ou
   Car == 95           .   %  um sublinhado ('_', ASCII = 95)


%===============================================================================
%Nome: escanear
%Descrio do predicado: escanea a lista de caracteres e monta a lista de tokens
%                        do programa em ALP sendo lido.
%
%Varivel de sada: 
%   - ListaTokens: a lista de tokens do programa em ALP lido.
%
%Assertiva de entrada:
%   - o arquivo com o programa ALP est aberto.
%===============================================================================

escanear( [ Token | ListaTokens ] )         :-
   tab( 4 )                                 ,   % imprime 4 brancos
   pegarCar( Car )                          ,   % pega o primeiro caractere
   pegarToken( Car, Token, NovoCar )        ,   % pega o primeiro token 
   restoProg( Token, NovoCar, ListaTokens ) ,   % escanea o resto dos tokens
   !                                        .


%===============================================================================
%Nome: pegarCar
%Descrio do predicado: pega o prximo caractere do cdigo fonte do programa em
%                        ALP sendo lido.
%
%Varivel de sada: 
%   - Car: um caractere presente no cdigo fonte do programa em ALP.
%
%Assertiva de entrada:
%   - o arquivo com o programa ALP est aberto.
%===============================================================================

% l um caractere, se houver erro, este  tratado; se no der erro e ele for fim 
% de linha, uma linha  pulada e 4 brancos so impressos; j se ele for fim de 
% arquivo, uma linha  pulada; por fim, se no for fim de linha ou de arquivo, o 
% imprime
pegarCar( Car )                                          :- 
   catch( get0( Car ), _Excecao, tratarErro( leitura ) ) ,  
   ( serFimLinha( Car )                                  ,   
     nl                                                  ,    
     tab( 4 )                                            ;    
     serFimArq( Car )                                    ,   
     nl                                                  ;  
     put( Car ) )                                        .   


%===============================================================================
%Nome: restoProg
%Descrio do predicado: escanea o restante da lista de caracteres e monta o 
%                        restante da lista de tokens do programa em ALP sendo 
%                        lido.
%
%Variveis de entrada: 
%   - UltToken: o ltimo token lido; e
%   - Car:      o ltimo caractere lido.
%
%Varivel de sada:
%   - ListaTokens: a lista de tokens do programa em ALP lido posteriores a
%                  UltToken.
%
%Assertiva de entrada:
%   - o arquivo com o programa ALP est aberto.
%===============================================================================

% se ler depois do fim de arquivo, no retorna tokens
restoProg( eop, _Car, [ ] ) .   

% se no ler depois do fim de arquivo, pega o prximo token, o coloca na lista
% de tokens e pega o restante da lista de tokens recursivamente
restoProg( _UltToken, Car, [ Token | ListaTokens ] ) :- 
   pegarToken( Car, Token, NovoCar )                 ,   
   restoProg( Token, NovoCar, ListaTokens )          .


%===============================================================================
%Nome: pegarToken
%Descrio do predicado: pega o prximo token do programa em ALP sendo lido.
%
%Varivel de entrada: 
%   - Car: o ltimo caractere lido.
%
%Variveis de sada:
%   - Token:     o token lido; e
%   - NovoCar_2: o caractere lido mais recentemente.
%
%Assertiva de entrada:
%   - o arquivo com o programa ALP est aberto.
%
%Assertiva de sada: 
%   - o arquivo manipulado continuar aberto a menos que um caractere invlido 
%     seja lido, o que far com que o programa seja abortado.
%===============================================================================

% se ler fim de arquivo, retorna o token correspondente 
pegarToken( Car, eop, 0 ) :-  
   serFimArq( Car )       .
  
% pega tokens com apenas um caractere 
pegarToken( Car, Token, NovoCar_2 ) :-
   serSimples( Car, Token )         ,
   pegarCar( NovoCar_2 )            .

% pega token que pode ser tanto simples ou duplo:
% se o primeiro caractere lido no fizer par com o prximo, o token  simples
% caso contrrio,  duplo
pegarToken( Car, Token, NovoCar_2 )  :-  
   serDuplo( Car, OutroToken )       ,
   pegarCar( NovoCar_1 )             ,
   ( serPar( Car, NovoCar_1, Token ) ,
     pegarCar( NovoCar_2 )           ; 
     Token = OutroToken              ,
     NovoCar_2 = NovoCar_1 )         .

% pega token que  um identificador, que tem como primeiro caractere uma letra
% o resto do token pode ser composto de letras, dgitos e sublinhados
% se o identificador for palavra reservada de ALP, apenas ele  retornado
% seno, ele  retornardo como um predicado que diz que ele  identificador
pegarToken( Car, Token, NovoCar_2 )               :-  
   ( serMinuscula( Car )                          ;
     serMaiuscula( Car ) )                        , 
   pegarCar( NovoCar_1 )                          , 
   pegarRestoId( NovoCar_1, ListaCar, NovoCar_2 ) ,   
   name( Id, [ Car | ListaCar ] )                 , 
   ( serReserv( Id )                              ,
     Token = Id                                   ; 
     Token = ide( Id ) )                          .

% pega token que  um nmero, o qual  composto apenas de dgitos
pegarToken( Car, num( Num ), NovoCar_2 )           :- 
   serDigito( Car )                                ,
   pegarCar( NovoCar_1 )                           ,
   pegarRestoNum( NovoCar_1, ListaCar, NovoCar_2 ) ,
   name( Num, [ Car | ListaCar ] )                 .

% se o ltimo caractere lido  um branco, apenas l um novo caractere
% e pega um token a partir dele
pegarToken( Car, Token, NovoCar_2 )        :-
   serBranco( Car )                        , 
   pegarCar( NovoCar )                     , 
   pegarToken( NovoCar, Token, NovoCar_2 ) .

% se o ltimo caractere lido  invlido, ou seja, no pode pertencer a nenhum
% token, isto  avisado ao usurio, e o programa  abortado, fechando antes o
% arquivo aberto
pegarToken( Car, _Token, _NovoCar_2 )  :-
   write( '\n\nCaractere ilegal: ' )   ,
   put( Car )                          ,
   nl                                  ,
   catch( seen, _Excecao, abort )      ,
   abort                               .


%===============================================================================
%Nome: pegarRestoId
%Descrio do predicado: pega o restante de um identificador.
%
%Varivel de entrada: 
%   - Car: o ltimo caractere lido.
%
%Variveis de sada:
%   - ListaCar:  a lista de caracteres lidos que representa um identificador; e
%   - NovoCar_2: o caractere lido mais recentemente.
%
%Assertiva de entrada:
%   - o arquivo com o programa ALP est aberto.
%===============================================================================

% se o ltimo caractere lido puder ser encontrado dentro de um identificador, a 
% busca por caracteres que possam pertencer a ele continua recursivamente
pegarRestoId( Car, [ Car | ListaCar ], NovoCar_2 ) :- 
   serCarVal( Car )                                , 
   pegarCar( NovoCar_1 )                           ,
   pegarRestoId( NovoCar_1, ListaCar, NovoCar_2 )  .

% se o ltimo caractere lido no puder ser encontrado dentro de um 
% identificador, pra a busca
pegarRestoId( Car, [ ], Car ) .


%===============================================================================
%Nome: pegarRestoNum
%Descrio do predicado: pega o restante de um nmero.
%
%Varivel de entrada: 
%   - Car: o ltimo caractere lido.
%
%Variveis de sada:
%   - ListaCar:  a lista de caracteres lidos que representa um nmero; e
%   - NovoCar_2: o caractere lido mais recentemente.
%
%Assertiva de entrada:
%   - o arquivo com o programa ALP est aberto.
%===============================================================================

% se o ltimo caractere lido for um dgito, a busca por dgitos continua 
% recursivamente
pegarRestoNum( Car, [ Car | ListaCar ], NovoCar_2 ) :- 
   serDigito( Car )                                 , 
   pegarCar( NovoCar_1 )                            , 
   pegarRestoNum( NovoCar_1, ListaCar, NovoCar_2 )  .

% se o ltimo caractere lido no for um dgito, pra a busca
pegarRestoNum( Car, [ ], Car ) .
